from typing import Generic, TypeVar

from ffun.core import errors

KEY = int | str


class Entity:
    __slots__ = ("key",)

    # TODO: refactor to correct type inference
    def __init__(self, key: KEY) -> None:
        self.key = key


ENTITY = TypeVar("ENTITY", bound=Entity)


class Register(Generic[ENTITY]):
    __slots__ = ("_entities",)

    def __init__(self) -> None:
        self._entities: dict[KEY, ENTITY] = {}

    def add(self, entity: ENTITY) -> None:
        if entity.key in self._entities:
            raise errors.EntityAlreadyRegistered()

        self._entities[entity.key] = entity

    def get(self, key: KEY) -> ENTITY:
        return self._entities[key]

    def has(self, key: KEY) -> bool:
        return key in self._entities

    def remove(self, key: KEY) -> None:
        del self._entities[key]

    def all(self) -> dict[KEY, ENTITY]:
        return dict(self._entities)



import psycopg

from ffun.core import logging
from ffun.core.postgresql import ExecuteType, execute, run_in_transaction, transaction
from ffun.domain.domain import new_user_id
from ffun.domain.entities import UserId
from ffun.users import errors
from ffun.users.entities import Service

logger = logging.get_module_logger()


async def store_user(execute: ExecuteType, internal_id: UserId) -> None:
    sql = """
         INSERT INTO u_users (id, created_at)
         VALUES (%(internal_id)s, NOW())
         ON CONFLICT (id) DO NOTHING
    """

    await execute(sql, {"internal_id": internal_id})


async def add_mapping(service: Service, external_id: str) -> UserId:
    sql = """
        INSERT INTO u_mapping (service_id, external_id, internal_id)
        VALUES (%(service_id)s, %(external_id)s, %(internal_id)s)
    """

    internal_id = new_user_id()

    try:
        async with transaction() as execute:
            await execute(sql, {"service_id": service, "external_id": external_id, "internal_id": internal_id})

            # must be after mapping registration, to avoid storing fake internal_id
            await store_user(execute, internal_id)

        logger.business_event("user_created", user_id=internal_id)

    except psycopg.errors.UniqueViolation:
        return await get_mapping(service, external_id)

    return internal_id


async def get_mapping(service: Service, external_id: str) -> UserId:
    sql = """
        SELECT internal_id
        FROM u_mapping
        WHERE service_id = %(service_id)s
              AND external_id = %(external_id)s
    """

    result = await execute(sql, {"service_id": service, "external_id": external_id})

    if not result:
        raise errors.NoUserMappingFound(service=service, external_id=external_id)

    return result[0]["internal_id"]  # type: ignore


async def get_user_external_ids(internal_id: UserId) -> dict[Service, str]:
    sql = """
    SELECT service_id, external_id
    FROM u_mapping
    WHERE internal_id = %(internal_id)s
    """

    result = await execute(sql, {"internal_id": internal_id})

    return {row["service_id"]: row["external_id"] for row in result}


async def unlink_user(service: Service, internal_id: UserId) -> None:
    sql = """
    DELETE FROM u_mapping
    WHERE service_id = %(service_id)s
          AND internal_id = %(internal_id)s
    """

    await execute(sql, {"service_id": service, "internal_id": internal_id})


async def count_total_users() -> int:
    result = await execute("SELECT COUNT(*) FROM u_users")
    return result[0]["count"]  # type: ignore


# Use only for development and testing purposes
@run_in_transaction
async def tech_move_user(execute: ExecuteType, from_user_id: UserId, to_user_id: UserId) -> None:
    await execute("DELETE FROM u_mapping WHERE internal_id = %(internal_id)s", {"internal_id": from_user_id})
    await execute(
        "UPDATE u_mapping SET internal_id = %(from_user_id)s WHERE internal_id = %(to_user_id)s",
        {"from_user_id": from_user_id, "to_user_id": to_user_id},
    )



"""
replace tokens traking with costs tracking
"""

from typing import Any

from psycopg import Connection
from psycopg.rows import dict_row
from yoyo import step

__depends__ = {"20230911_01_5vjXI-index-to-search-by-value"}

sql = """
INSERT INTO us_settings (user_id, kind, value)
VALUES (%(user_id)s, 6, %(value)s)
"""


cost_1m = 0.6


def tokens_to_costs(tokens: int) -> int:
    return int(tokens / 1_000_000 * cost_1m)


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor(row_factory=dict_row)

    cursor.execute("SELECT * FROM us_settings WHERE kind=2")

    for row in cursor.fetchall():

        cursor.execute(
            sql,
            {
                "user_id": row["user_id"],
                "value": tokens_to_costs(int(row["value"])),
            },
        )

    cursor.execute("DELETE FROM us_settings WHERE kind=2")


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    pass


steps = [step(apply_step, rollback_step)]



"""
source-table
"""

import re
import unicodedata
from typing import Any

from furl import furl
from psycopg import Connection
from psycopg.rows import dict_row
from yoyo import step

__depends__ = {"20240512_01_jL7Mt-turn-on-unique-uids", "20230427_01_9HuHj-add-feeds-links"}


RE_SCHEMA = re.compile(r"^(\w+):")


def _fake_schema_for_url(url: str) -> str:
    url = url.strip()

    if url.startswith("//"):
        return url

    if RE_SCHEMA.match(url):
        return url

    if "." not in url.split("/")[0]:
        return url

    return f"//{url}"


def url_to_source_uid(url: str) -> str:
    normalized_url = unicodedata.normalize("NFC", url).lower().strip()

    normalized_url = _fake_schema_for_url(normalized_url)

    url_object = furl(normalized_url)

    domain = url_object.host

    if domain.startswith("www."):
        domain = domain[4:]

    if domain.endswith(".reddit.com"):
        domain = "reddit.com"

    assert isinstance(domain, str)

    return domain


sql_sources_table = """
CREATE TABLE f_sources (
    id UUID PRIMARY KEY,
    uid TEXT NOT NULL UNIQUE,
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW()
)
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor(row_factory=dict_row)

    ####################################
    # cleaning up feeds from broken urls
    cursor.execute("SELECT id FROM f_feeds WHERE url NOT LIKE 'http%'")
    result = cursor.fetchall()
    feed_ids = [row["id"] for row in result]
    cursor.execute("DELETE FROM fl_links WHERE feed_id = ANY(%(feed_ids)s)", {"feed_ids": feed_ids})
    cursor.execute("DELETE FROM f_feeds WHERE id = ANY(%(feed_ids)s)", {"feed_ids": feed_ids})
    ####################################

    cursor.execute(sql_sources_table)

    cursor.execute("ALTER TABLE f_feeds ADD COLUMN source_id UUID REFERENCES f_sources (id) NULL")

    cursor.execute("SELECT * from f_feeds")

    result = cursor.fetchall()

    for row in result:
        source_uid = url_to_source_uid(row["url"])

        cursor.execute(
            "INSERT INTO f_sources (id, uid) VALUES (gen_random_uuid(), %(uid)s) ON CONFLICT DO NOTHING",
            {"uid": source_uid},
        )

        cursor.execute(
            """UPDATE f_feeds
                          SET source_id = (SELECT id FROM f_sources WHERE uid = %(source_uid)s)
                          WHERE id = %(feed_id)s""",
            {"source_uid": source_uid, "feed_id": row["id"]},
        )

    cursor.execute("ALTER TABLE f_feeds ALTER COLUMN source_id SET NOT NULL")


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()

    cursor.execute("ALTER TABLE f_feeds DROP COLUMN source_id")
    cursor.execute("DROP TABLE f_sources")


steps = [step(apply_step, rollback_step)]



import uuid
from typing import Iterable, Sequence

from bidict import bidict
from pypika import PostgreSQLQuery

from ffun.core import logging
from ffun.core.postgresql import ExecuteType, execute
from ffun.domain.entities import EntryId, TagId, TagUid
from ffun.ontology import errors
from ffun.ontology.entities import TagCategory, TagProperty, TagPropertyType

logger = logging.get_module_logger()


async def get_tags_mappig() -> bidict[TagUid, TagId]:
    sql = "SELECT id, uid FROM o_tags"
    rows = await execute(sql)
    return bidict.bidict({row["uid"]: row["id"] for row in rows})  # type: ignore


async def get_id_by_tag(tag: TagUid) -> TagId | None:
    sql = "SELECT id FROM o_tags WHERE uid = %(tag)s"
    rows = await execute(sql, {"tag": tag})

    if not rows:
        return None

    return rows[0]["id"]  # type: ignore


async def register_tag(tag: TagUid) -> TagId:
    sql = """
    INSERT INTO o_tags (uid)
    VALUES (%(tag)s)
    ON CONFLICT (uid) DO NOTHING
    RETURNING id"""

    rows = await execute(sql, {"tag": tag})

    if not rows:
        tag_id = await get_id_by_tag(tag)

        assert tag_id is not None
        assert isinstance(tag_id, int)

        return tag_id

    assert isinstance(rows[0]["id"], int)

    return rows[0]["id"]  # type: ignore


async def get_or_create_id_by_tag(tag: TagUid) -> TagId:
    tag_id = await get_id_by_tag(tag)

    if tag_id is not None:
        return tag_id

    return await register_tag(tag)


async def get_tags_by_ids(tags_ids: list[TagId]) -> dict[TagId, TagUid]:
    sql = "SELECT * FROM o_tags WHERE id = ANY(%(tags_ids)s)"
    rows = await execute(sql, {"tags_ids": tags_ids})
    return {row["id"]: row["uid"] for row in rows}


async def _save_tags(execute: ExecuteType, entry_id: EntryId, tags_ids: Iterable[TagId]) -> None:
    if not tags_ids:
        return

    query = PostgreSQLQuery.into("o_relations").columns("entry_id", "tag_id")

    for tag_id in tags_ids:
        query = query.insert(entry_id, tag_id)

    query = query.on_conflict("entry_id", "tag_id").do_nothing()

    await execute(str(query))


async def _register_relations_processors(
    execute: ExecuteType, relations_ids: Iterable[int], processor_id: int
) -> None:

    if not relations_ids:
        return

    query = PostgreSQLQuery.into("o_relations_processors").columns("relation_id", "processor_id")

    for relation_id in relations_ids:
        query = query.insert(relation_id, processor_id)

    query = query.on_conflict("relation_id", "processor_id").do_nothing()

    await execute(str(query))


async def _get_relations_for_entry_and_tags(
    execute: ExecuteType, entry_id: EntryId, tags_ids: Iterable[TagId]
) -> dict[TagId, int]:
    result = await execute(
        "SELECT id, tag_id FROM o_relations WHERE entry_id = %(entry_id)s AND tag_id = ANY(%(tags_ids)s)",
        {"entry_id": entry_id, "tags_ids": list(tags_ids)},
    )

    return {row["tag_id"]: row["id"] for row in result}


async def apply_tags(execute: ExecuteType, entry_id: EntryId, processor_id: int, tags_ids: Iterable[TagId]) -> None:
    await _save_tags(execute, entry_id, tags_ids)

    relations = await _get_relations_for_entry_and_tags(execute, entry_id, tags_ids)

    await _register_relations_processors(execute, list(relations.values()), processor_id)


async def tech_copy_relations(execute: ExecuteType, entry_from_id: EntryId, entry_to_id: EntryId) -> None:
    """Copy relations with processors info."""
    # get processors for each tag in entry_from
    sql = """
    SELECT rp.processor_id, r.tag_id
    FROM o_relations_processors rp
    JOIN o_relations r ON rp.relation_id = r.id
    WHERE r.entry_id = %(entry_id)s
    """

    result = await execute(sql, {"entry_id": entry_from_id})

    tags_by_processors: dict[int, list[TagId]] = {}

    for row in result:
        if row["processor_id"] not in tags_by_processors:
            tags_by_processors[row["processor_id"]] = []

        tags_by_processors[row["processor_id"]].append(row["tag_id"])

    for processor_id, tags_ids in tags_by_processors.items():
        await apply_tags(execute, entry_to_id, processor_id, tags_ids)


async def apply_tags_properties(execute: ExecuteType, properties: Sequence[TagProperty]) -> None:

    if not properties:
        return

    # check for duplicates
    tags_ids = {(property.tag_id, property.type, property.processor_id) for property in properties}

    if len(tags_ids) != len(properties):
        raise errors.DuplicatedTagPropeties()

    query = PostgreSQLQuery.into("o_tags_properties").columns("tag_id", "type", "value", "processor_id", "created_at")

    # sort properties to avoid deadlocks
    properties = sorted(properties, key=lambda p: (p.tag_id, p.type, p.processor_id))

    for property in properties:
        query = query.insert(
            property.tag_id,
            property.type,
            property.value,
            property.processor_id,
            property.created_at,
        )

    query = query.on_conflict("tag_id", "type", "processor_id").do_update("value")

    await execute(str(query))


async def get_tags_for_entries(execute: ExecuteType, entries_ids: list[EntryId]) -> dict[EntryId, set[TagId]]:
    sql = """SELECT CONCAT(entry_id::text, '|', tag_id::text) AS ids
             FROM o_relations
             WHERE entry_id = ANY(%(entries_ids)s)"""

    rows = await execute(sql, {"entries_ids": entries_ids})

    result: dict[EntryId, set[TagId]] = {}

    entry_ids_mapping: dict[str, EntryId] = {}

    for row in rows:
        raw_entry_id, raw_tag_id = row["ids"].split("|")

        if raw_entry_id not in entry_ids_mapping:
            entry_ids_mapping[raw_entry_id] = EntryId(uuid.UUID(raw_entry_id))

        tag_id = int(raw_tag_id)

        entry_id = entry_ids_mapping[raw_entry_id]

        if entry_id not in result:
            result[entry_id] = set()

        result[entry_id].add(tag_id)  # type: ignore

    return result


async def get_tags_properties(tags_ids: Iterable[TagId]) -> list[TagProperty]:
    result = await execute(
        "SELECT * FROM o_tags_properties WHERE tag_id = ANY(%(tags_ids)s) ORDER BY created_at DESC",
        {"tags_ids": list(tags_ids)},
    )

    return [
        TagProperty(
            tag_id=row["tag_id"],
            type=TagPropertyType(row["type"]),
            value=row["value"],
            processor_id=row["processor_id"],
            created_at=row["created_at"],
        )
        for row in result
    ]


async def remove_relations_for_entries(execute: ExecuteType, entries_ids: list[EntryId]) -> None:
    sql = "SELECT id FROM o_relations WHERE entry_id = ANY(%(entries_ids)s)"

    result = await execute(sql, {"entries_ids": entries_ids})

    relations_ids = [row["id"] for row in result]

    sql = "DELETE FROM o_relations_processors WHERE relation_id = ANY(%(relations_ids)s)"

    await execute(sql, {"relations_ids": relations_ids})

    sql = "DELETE FROM o_relations WHERE entry_id = ANY(%(entries_ids)s)"

    await execute(sql, {"entries_ids": entries_ids})


async def count_total_tags() -> int:
    result = await execute("SELECT COUNT(*) FROM o_tags")
    return result[0]["count"]  # type: ignore


async def count_total_tags_per_category() -> dict[TagCategory, int]:

    numbers: dict[TagCategory, int] = {}

    sql = """
    SELECT count(*)
    FROM o_tags_properties
    WHERE type = %(type)s AND value LIKE %(value)s
    """

    for category in TagCategory:
        result = await execute(sql, {"type": TagPropertyType.categories, "value": f"%{category.value}%"})

        numbers[category] = result[0]["count"]

    return numbers


async def count_total_tags_per_type() -> dict[TagPropertyType, int]:

    numbers: dict[TagPropertyType, int] = {type_: 0 for type_ in TagPropertyType}

    sql = """
    SELECT type, count(*)
    FROM o_tags_properties
    GROUP BY type
    """

    result = await execute(sql)

    for row in result:
        numbers[TagPropertyType(row["type"])] = row["count"]

    return numbers



# TODO: rename this module, because mypy does not support name `types`
# TODO: refactor to use pydantic instead of custom types

import decimal
import enum
from typing import Any

from cryptography.fernet import Fernet

from ffun.user_settings import errors
from ffun.user_settings.settings import settings


class TypeId(str, enum.Enum):
    integer = "integer"
    string = "string"
    boolean = "boolean"
    secret = "secret"  # noqa: S105
    decimal = "decimal"


class Type:
    id: TypeId = NotImplemented

    def serialize(self, value: Any) -> str:
        raise NotImplementedError('You must implement "serialize" method in child class')

    def deserialize(self, data: str) -> Any:
        raise NotImplementedError('You must implement "deserialize" method in child class')

    def normalize(self, value: Any) -> Any:
        raise NotImplementedError('You must implement "normalize" method in child class')


class Integer(Type):
    id = TypeId.integer

    def serialize(self, value: int) -> str:
        if not isinstance(value, int):
            raise errors.WrongValueType(value=value)

        return str(value)

    def deserialize(self, data: str) -> int:
        return int(data)

    def normalize(self, value: Any) -> int:
        if value == "":
            return 0

        return int(value)


# TODO: test
class Decimal(Type):
    id = TypeId.decimal

    def serialize(self, value: decimal.Decimal) -> str:
        if not isinstance(value, decimal.Decimal):
            raise errors.WrongValueType(value=value)

        return str(value)

    def deserialize(self, data: str) -> decimal.Decimal:
        return decimal.Decimal(data)

    def normalize(self, value: Any) -> decimal.Decimal:
        if value == "":
            return decimal.Decimal(0)

        # Not the best place to apply normalization of such a kind
        # But for now, it is the simplest way to handle it
        # TODO: move decimal normalization to the frontend side, when the frontend is ready
        if isinstance(value, str):
            value = value.strip().replace(",", ".")

        return decimal.Decimal(value)


class String(Type):
    id = TypeId.string

    def serialize(self, value: str) -> str:
        return value

    def deserialize(self, data: str) -> str:
        return data

    def normalize(self, value: Any) -> str:
        return str(value)


class Boolean(Type):
    id = TypeId.boolean

    def serialize(self, value: bool) -> str:
        return "true" if value else "false"

    def deserialize(self, data: str) -> bool:
        return data == "true"

    def normalize(self, value: Any) -> bool:
        return value in (True, "true", "True")


class Secret(Type):
    id = TypeId.secret

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.fernet = Fernet(settings.secret_key)

    def serialize(self, value: str) -> str:
        return self.fernet.encrypt(value.encode()).decode()

    def deserialize(self, data: str) -> str:
        return self.fernet.decrypt(data.encode()).decode()

    def normalize(self, value: Any) -> str:
        return str(value)



import contextlib
from typing import Any, AsyncGenerator

from fastapi import FastAPI
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.asyncio import delete_user
from supertokens_python.framework.fastapi import get_middleware
from supertokens_python.ingredients.emaildelivery.types import EmailDeliveryConfig
from supertokens_python.recipe import dashboard, emailverification, passwordless, session, usermetadata
from supertokens_python.recipe.passwordless import ContactEmailOnlyConfig
from supertokens_python.recipe.passwordless.types import EmailDeliveryOverrideInput, EmailTemplateVars
from supertokens_python.recipe.session.asyncio import revoke_all_sessions_for_user

from ffun.auth.settings import settings
from ffun.core import logging

logger = logging.get_module_logger()


def custom_email_deliver(original_implementation: EmailDeliveryOverrideInput) -> EmailDeliveryOverrideInput:
    original_send_email = original_implementation.send_email

    async def send_email(template_vars: EmailTemplateVars, user_context: dict[str, Any]) -> None:
        assert template_vars.url_with_link_code is not None
        # By default: `${websiteDomain}/${websiteBasePath}/verify`
        template_vars.url_with_link_code = template_vars.url_with_link_code.replace("/auth/verify", "/auth")
        return await original_send_email(template_vars, user_context)

    original_implementation.send_email = send_email  # type: ignore
    return original_implementation


@contextlib.asynccontextmanager
async def use_supertokens(
    app_name: str,
    api_domain: str,
    website_domain: str,
    cookie_secure: bool = settings.supertokens.cookie_secure,
    api_base_path: str = settings.supertokens.api_base_path,
    website_base_path: str = settings.supertokens.website_base_path,
) -> AsyncGenerator[None, None]:
    logger.info("supertokens_enabled")

    init(
        app_info=InputAppInfo(
            app_name=app_name,
            api_domain=api_domain,
            website_domain=website_domain,
            api_base_path=api_base_path,
            website_base_path=website_base_path,
        ),
        supertokens_config=SupertokensConfig(
            connection_uri=settings.supertokens.connection_uri, api_key=settings.supertokens.api_key
        ),
        framework="fastapi",
        recipe_list=[
            emailverification.init(mode="REQUIRED"),
            session.init(cookie_secure=cookie_secure),
            passwordless.init(
                flow_type="MAGIC_LINK",
                contact_config=ContactEmailOnlyConfig(),
                email_delivery=EmailDeliveryConfig(override=custom_email_deliver),
            ),
            dashboard.init(),
            usermetadata.init(),
        ],
        mode="asgi",
    )

    logger.info("supertokens_initialized")

    yield


def add_middlewares(app: FastAPI) -> None:
    app.add_middleware(get_middleware())


async def remove_user(user_id: str) -> None:
    logger.info("removing_supertokens_user", user_id=user_id)
    await delete_user(user_id=user_id)
    logger.info("supertokens_user_removed")


async def logout_user_from_all_sessions(user_id: str) -> None:
    logger.info("logging_out_supertokens_user", user_id=user_id)
    await revoke_all_sessions_for_user(user_id=user_id)
    logger.info("supertokens_user_logged_out")



import datetime

import pytest

from ffun.domain.datetime_intervals import month_interval_start


class TestMonthIntervalStart:

    @pytest.mark.asyncio
    async def test_month_interval_start(self) -> None:
        now = datetime.datetime.now()

        assert month_interval_start(now) == datetime.datetime(
            year=now.year,
            month=now.month,
            day=1,
            hour=0,
            minute=0,
            second=0,
            microsecond=0,
            tzinfo=now.tzinfo,
        )



from decimal import Decimal

import pytest
from pytest_mock import MockerFixture

from ffun.llms_framework import errors
from ffun.llms_framework.entities import (
    KeyStatus,
    LLMApiKey,
    LLMConfiguration,
    LLMProvider,
    LLMTokens,
    ModelInfo,
    USDCost,
)
from ffun.llms_framework.provider_interface import ProviderTest


class TestBaseProviderInterfaceClass:
    """Test concrete methods of the abstract ProviderInterface class."""

    def test_different_storages_for_keys(self, fake_llm_api_key: LLMApiKey) -> None:
        provider_1 = ProviderTest()
        provider_2 = ProviderTest()

        assert provider_1.api_keys_statuses is not provider_2.api_keys_statuses

        provider_1.api_keys_statuses.set(fake_llm_api_key, KeyStatus.works)
        provider_2.api_keys_statuses.set(fake_llm_api_key, KeyStatus.broken)

        assert provider_1.api_keys_statuses.get(fake_llm_api_key) == KeyStatus.works
        assert provider_2.api_keys_statuses.get(fake_llm_api_key) == KeyStatus.broken

    def test_get_model(self, fake_llm_provider: ProviderTest, mocker: MockerFixture) -> None:
        config_1 = LLMConfiguration(
            model="test-model-1",
            system="system prompt",
            max_return_tokens=LLMTokens(143),
            text_parts_intersection=100,
            temperature=0,
            top_p=0,
            presence_penalty=0,
            frequency_penalty=0,
        )

        assert fake_llm_provider.get_model(config_1) == ModelInfo(
            provider=LLMProvider.test,
            name="test-model-1",
            max_context_size=LLMTokens(12800),
            max_return_tokens=LLMTokens(4096),
            max_tokens_per_entry=LLMTokens(300000),
            input_1m_tokens_cost=USDCost(Decimal("0.3")),
            output_1m_tokens_cost=USDCost(Decimal("0.7")),
        )

        config_2 = config_1.replace(model="test-model-2")

        assert fake_llm_provider.get_model(config_2) == ModelInfo(
            provider=LLMProvider.test,
            name="test-model-2",
            max_context_size=LLMTokens(14212),
            max_return_tokens=LLMTokens(1024),
            max_tokens_per_entry=LLMTokens(300000),
            input_1m_tokens_cost=USDCost(Decimal("0.7")),
            output_1m_tokens_cost=USDCost(Decimal("0.3")),
        )

        mocker.patch.object(fake_llm_provider, "provider", LLMProvider.openai)

        config_3 = config_1.replace(provider=LLMProvider.openai, model="gpt-4o")

        assert fake_llm_provider.get_model(config_3) == ModelInfo(
            provider=LLMProvider.openai,
            name="gpt-4o",
            max_context_size=LLMTokens(128000),
            max_return_tokens=LLMTokens(4096),
            max_tokens_per_entry=LLMTokens(300000),
            input_1m_tokens_cost=USDCost(Decimal("5")),
            output_1m_tokens_cost=USDCost(Decimal("15")),
        )

    def test_wrong_provider(self, fake_llm_provider: ProviderTest, mocker: MockerFixture) -> None:

        config = LLMConfiguration(
            model="gpt-4o",
            system="system prompt",
            max_return_tokens=LLMTokens(143),
            text_parts_intersection=100,
            temperature=0,
            top_p=0,
            presence_penalty=0,
            frequency_penalty=0,
        )

        with pytest.raises(errors.ModelDoesNotFound):
            fake_llm_provider.get_model(config)

        mocker.patch.object(fake_llm_provider, "provider", LLMProvider.openai)

        assert fake_llm_provider.get_model(config) is not None

    def test_wrong_model(self, fake_llm_provider: ProviderTest, mocker: MockerFixture) -> None:
        config_1 = LLMConfiguration(
            model="test-model-wrong",
            system="system prompt",
            max_return_tokens=LLMTokens(143),
            text_parts_intersection=100,
            temperature=0,
            top_p=0,
            presence_penalty=0,
            frequency_penalty=0,
        )

        with pytest.raises(errors.ModelDoesNotFound):
            fake_llm_provider.get_model(config_1)

        config_2 = config_1.replace(model="test-model-1")

        assert fake_llm_provider.get_model(config_2) is not None



import pytest

from ffun.parsers.tests.helpers import feeds_fixtures_directory, feeds_fixtures_names


@pytest.fixture
def raw_feed_content() -> str:
    fixture_filename = feeds_fixtures_names()[0]

    with open(feeds_fixtures_directory / fixture_filename, encoding="utf-8") as f:
        return f.read()


@pytest.fixture
def another_raw_feed_content() -> str:
    fixture_filename = feeds_fixtures_names()[1]

    with open(feeds_fixtures_directory / fixture_filename, encoding="utf-8") as f:
        return f.read()



from ffun.parsers import feed, opml
from ffun.parsers.entities import FeedInfo


def parse_opml(data: str) -> list[FeedInfo]:
    return opml.extract_feeds(data)


create_opml = opml.create_opml


parse_feed = feed.parse_feed



"""
processors_pointers
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__: set[str] = set()


sql_create_processor_pointers = """
CREATE TABLE ln_processor_pointers (
    processor_id INTEGER PRIMARY KEY,
    pointer_created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT '1970-01-01 00:00:00',
    pointer_entry_id UUID NOT NULL DEFAULT '00000000-0000-0000-0000-000000000000',
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT CURRENT_TIMESTAMP,
    updated_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT CURRENT_TIMESTAMP
)
"""

sql_create_processors_queue = """
CREATE TABLE ln_processors_queue (
    id UUID PRIMARY KEY DEFAULT gen_random_uuid(),
    processor_id INTEGER NOT NULL,
    entry_id UUID NOT NULL,
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT CURRENT_TIMESTAMP,

    UNIQUE (processor_id, entry_id)
)
"""


sql_create_processors_queue_index = """
CREATE INDEX idx_ln_processors_queue_processor_id_created_at ON ln_processors_queue (processor_id, created_at)
"""


sql_create_processors_failed_entries = """
CREATE TABLE ln_failed_entries (
    id UUID PRIMARY KEY DEFAULT gen_random_uuid(),
    processor_id INTEGER NOT NULL,
    entry_id UUID NOT NULL,
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT CURRENT_TIMESTAMP,

    UNIQUE (processor_id, entry_id)
)
"""

sql_create_processors_failed_entries_index = """
CREATE INDEX idx_ln_failed_entries_processor_id_created_at ON ln_failed_entries (processor_id, created_at)
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()

    cursor.execute(sql_create_processor_pointers)

    cursor.execute(sql_create_processors_queue)
    cursor.execute(sql_create_processors_queue_index)

    cursor.execute(sql_create_processors_failed_entries)
    cursor.execute(sql_create_processors_failed_entries_index)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("DROP TABLE ln_processor_pointers, ln_processors_queue, ln_failed_entries")


steps = [step(apply_step, rollback_step)]



import datetime

import pydantic
import pydantic_settings

from ffun.core.settings import BaseSettings


class Proxy(pydantic.BaseModel):
    name: str
    url: str | None = None


class Settings(BaseSettings):
    loaders_number: int = 5
    max_concurrent_http_requests: int = 10
    minimum_period: datetime.timedelta = datetime.timedelta(hours=1)
    proxies: list[Proxy] = [Proxy(name="default", url=None)]

    proxy_anchors: list[str] = ["https://www.google.com", "https://www.amazon.com"]
    proxy_available_check_period: datetime.timedelta = datetime.timedelta(minutes=5)

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="FFUN_LOADER_")


settings = Settings()



"""
entries_table
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__: set[str] = set()


sql_create_entries_table = """
CREATE TABLE l_entries (
    id UUID PRIMARY KEY,
    feed_id UUID NOT NULL,
    title TEXT NOT NULL,
    body TEXT NOT NULL,
    external_id TEXT NOT NULL UNIQUE,
    external_url TEXT NOT NULL,
    external_tags TEXT[] NOT NULL,
    published_at TIMESTAMP WITH TIME ZONE NOT NULL,
    cataloged_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW(),
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW()
)
"""


sql_create_feeds_index = """
CREATE INDEX idx_l_entries_feed_id ON l_entries (feed_id);
"""


sql_create_entry_process_info_table = """
CREATE TABLE l_entry_process_info (
    entry_id UUID NOT NULL REFERENCES l_entries (id),
    processor_id INTEGER NOT NULL,
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW(),
    updated_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW(),
    processed_at TIMESTAMP WITH TIME ZONE NULL,

    PRIMARY KEY (processor_id, entry_id)
)
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute(sql_create_entries_table)
    cursor.execute(sql_create_feeds_index)
    cursor.execute(sql_create_entry_process_info_table)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("DROP TABLE l_entries, l_entry_process_info")


steps = [step(apply_step, rollback_step)]



import uuid

import pytest
import pytest_asyncio

from ffun.domain.entities import UserId
from ffun.users.domain import get_or_create_user_id
from ffun.users.entities import Service


@pytest.fixture
def external_user_id() -> str:
    return f"external-user-{uuid.uuid4().hex}"


@pytest.fixture
def another_external_user_id() -> str:
    return f"another-external-user-{uuid.uuid4().hex}"


@pytest_asyncio.fixture
async def internal_user_id(external_user_id: str) -> UserId:
    return await get_or_create_user_id(Service.supertokens, external_user_id)


@pytest_asyncio.fixture
async def another_internal_user_id(another_external_user_id: str) -> UserId:
    return await get_or_create_user_id(Service.supertokens, another_external_user_id)


@pytest.fixture
def five_external_user_ids() -> list[str]:
    return [f"external-user-{uuid.uuid4().hex}" for _ in range(5)]


@pytest_asyncio.fixture
async def five_internal_user_ids(five_external_user_ids: list[str]) -> list[UserId]:
    return [
        await get_or_create_user_id(Service.supertokens, external_user_id)
        for external_user_id in five_external_user_ids
    ]



from typing import Iterable

from ffun.core import logging
from ffun.core.postgresql import execute
from ffun.domain.entities import UserId
from ffun.user_settings.entities import UserSettings

logger = logging.get_module_logger()


async def save_setting(user_id: UserId, kind: int, value: str) -> None:
    sql = """
        INSERT INTO us_settings (user_id, kind, value)
        VALUES (%(user_id)s, %(kind)s, %(value)s)
        ON CONFLICT (user_id, kind)
        DO UPDATE SET value = %(value)s,
                      updated_at = NOW()
        RETURNING created_at, updated_at
    """

    results = await execute(sql, {"user_id": user_id, "kind": kind, "value": value})

    first_set = results[0]["created_at"] == results[0]["updated_at"]

    logger.business_event("setting_updated", user_id=user_id, kind=kind, first_set=first_set)


async def load_settings_for_users(user_ids: Iterable[UserId], kinds: Iterable[int]) -> dict[UserId, UserSettings]:
    sql = """
        SELECT *
        FROM us_settings
        WHERE user_id = ANY(%(user_ids)s)
        AND kind = ANY(%(kinds)s)
    """

    result = await execute(sql, {"user_ids": list(user_ids), "kinds": list(kinds)})

    values: dict[UserId, UserSettings] = {user_id: {} for user_id in user_ids}

    for row in result:
        user_id = row["user_id"]
        kind = row["kind"]
        value = row["value"]

        values[user_id][kind] = value

    return values


async def get_users_with_setting(kind: int, value: str) -> set[UserId]:
    sql = """
        SELECT user_id
        FROM us_settings
        WHERE kind = %(kind)s
        AND value = %(value)s
    """

    result = await execute(sql, {"kind": kind, "value": value})

    return {row["user_id"] for row in result}


async def remove_setting_for_all_users(kind: int) -> None:
    sql = """
        DELETE FROM us_settings
        WHERE kind = %(kind)s
    """

    await execute(sql, {"kind": kind})


async def find_all_kinds() -> set[int]:
    sql = """
        SELECT kind
        FROM us_settings
        GROUP by kind
    """

    result = await execute(sql)
    return {row["kind"] for row in result}



import asyncio
import datetime
from typing import Any

from ffun.core import logging, utils
from ffun.core.background_tasks import InfiniteTask
from ffun.feeds import domain as f_domain
from ffun.loader import domain
from ffun.loader.settings import settings

logger = logging.get_module_logger()


# TODO: tests
class FeedsLoader(InfiniteTask):
    __slots__ = ("_loaders_number", "_last_proxies_check")

    def __init__(self, loaders_number: int = settings.loaders_number, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._loaders_number = loaders_number
        self._last_proxies_check = datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc)

    async def single_run(self) -> None:
        if utils.now() - self._last_proxies_check > settings.proxy_available_check_period:
            await domain.check_proxies_availability()
            self._last_proxies_check = utils.now()

        feeds = await f_domain.get_next_feeds_to_load(
            number=self._loaders_number, loaded_before=datetime.datetime.utcnow() - settings.minimum_period
        )

        tasks = [domain.process_feed(feed=feed) for feed in feeds]

        await asyncio.gather(*tasks, return_exceptions=True)



import uuid

import pytest

from ffun.core.postgresql import execute, transaction
from ffun.core.tests.helpers import Delta, TableSizeDelta, TableSizeNotChanged
from ffun.domain.entities import EntryId, TagId, TagUid
from ffun.library.entities import Entry
from ffun.ontology import errors
from ffun.ontology.domain import apply_tags_to_entry
from ffun.ontology.entities import ProcessorTag, TagCategory, TagPropertyType
from ffun.ontology.operations import (
    _get_relations_for_entry_and_tags,
    _register_relations_processors,
    _save_tags,
    apply_tags,
    apply_tags_properties,
    count_total_tags,
    count_total_tags_per_category,
    count_total_tags_per_type,
    get_or_create_id_by_tag,
    get_tags_properties,
    remove_relations_for_entries,
    tech_copy_relations,
)
from ffun.ontology.tests.helpers import assert_has_tags


async def assert_tags_processors(entry_id: EntryId, tag_processors: dict[TagId, set[int]]) -> None:
    relations = await _get_relations_for_entry_and_tags(execute, entry_id, list(tag_processors.keys()))

    relations_ids = list(relations.values())

    sql = "SELECT relation_id, processor_id FROM o_relations_processors WHERE relation_id = ANY(%(relations_ids)s)"

    result = await execute(sql, {"relations_ids": relations_ids})

    expected = set()

    for tag_id, processor_ids in tag_processors.items():
        for processor_id in processor_ids:
            expected.add((relations[tag_id], processor_id))

    assert expected == {(row["relation_id"], row["processor_id"]) for row in result}


class TestSaveTags:
    @pytest.mark.asyncio
    async def test_no_tags(self, cataloged_entry: Entry) -> None:
        async with TableSizeNotChanged("o_relations"):
            await _save_tags(execute, cataloged_entry.id, [])

    @pytest.mark.asyncio
    async def test(
        self, cataloged_entry: Entry, another_cataloged_entry: Entry, three_tags_ids: tuple[TagId, TagId, TagId]
    ) -> None:
        async with TableSizeDelta("o_relations", delta=4):
            await _save_tags(execute, cataloged_entry.id, three_tags_ids[:2])
            await _save_tags(execute, another_cataloged_entry.id, three_tags_ids[1:])

        await assert_has_tags(
            {
                cataloged_entry.id: {three_tags_ids[0], three_tags_ids[1]},
                another_cataloged_entry.id: {three_tags_ids[1], three_tags_ids[2]},
            }
        )


class TestRegisterRelationsProcessors:
    @pytest.mark.asyncio
    async def test_no_relations(self, fake_processor_id: int) -> None:
        async with TableSizeNotChanged("o_relations_processors"):
            await _register_relations_processors(execute, [], fake_processor_id)

    @pytest.mark.asyncio
    async def test(
        self, cataloged_entry: Entry, fake_processor_id: int, three_tags_ids: tuple[TagId, TagId, TagId]
    ) -> None:
        await _save_tags(execute, cataloged_entry.id, three_tags_ids[:2])

        relations = await _get_relations_for_entry_and_tags(execute, cataloged_entry.id, three_tags_ids[:2])

        async with TableSizeDelta("o_relations_processors", delta=2):
            await _register_relations_processors(
                execute, relations_ids=list(relations.values()), processor_id=fake_processor_id
            )

        await assert_tags_processors(
            entry_id=cataloged_entry.id,
            tag_processors={three_tags_ids[0]: {fake_processor_id}, three_tags_ids[1]: {fake_processor_id}},
        )

    @pytest.mark.asyncio
    async def test_duplicated(
        self, cataloged_entry: Entry, fake_processor_id: int, three_tags_ids: tuple[TagId, TagId, TagId]
    ) -> None:
        await _save_tags(execute, cataloged_entry.id, three_tags_ids)

        relations = await _get_relations_for_entry_and_tags(execute, cataloged_entry.id, three_tags_ids)

        relation_ids = list(relations.values())

        async with TableSizeDelta("o_relations_processors", delta=3):
            await _register_relations_processors(
                execute, relations_ids=relation_ids[:2], processor_id=fake_processor_id
            )
            await _register_relations_processors(
                execute, relations_ids=relation_ids[1:], processor_id=fake_processor_id
            )

        await assert_tags_processors(
            entry_id=cataloged_entry.id,
            tag_processors={
                three_tags_ids[0]: {fake_processor_id},
                three_tags_ids[1]: {fake_processor_id},
                three_tags_ids[2]: {fake_processor_id},
            },
        )


class TestGetRelationsForEntryAndTags:
    """Tested in other tests & code."""


class TestTechCopyRelations:
    @pytest.mark.asyncio
    async def test_no_relations(self, cataloged_entry: Entry, another_cataloged_entry: Entry) -> None:
        async with TableSizeNotChanged("o_relations"):
            await tech_copy_relations(execute, cataloged_entry.id, another_cataloged_entry.id)

    @pytest.mark.asyncio
    async def test_full_copy(
        self,
        cataloged_entry: Entry,
        another_cataloged_entry: Entry,
        fake_processor_id: int,
        another_fake_processor_id: int,
        three_tags_ids: tuple[TagId, TagId, TagId],
    ) -> None:
        async with transaction() as trx:
            await apply_tags(
                trx, entry_id=cataloged_entry.id, processor_id=fake_processor_id, tags_ids=three_tags_ids[:2]
            )

        async with transaction() as trx:
            await apply_tags(
                trx, entry_id=cataloged_entry.id, processor_id=another_fake_processor_id, tags_ids=three_tags_ids[1:]
            )

        async with (
            TableSizeDelta("o_relations_processors", delta=4),
            TableSizeDelta("o_relations", delta=3),
            TableSizeNotChanged("o_tags"),
        ):
            async with transaction() as trx:
                await tech_copy_relations(trx, cataloged_entry.id, another_cataloged_entry.id)

        await assert_has_tags({another_cataloged_entry.id: set(three_tags_ids)})

        await assert_tags_processors(
            entry_id=another_cataloged_entry.id,
            tag_processors={
                three_tags_ids[0]: {fake_processor_id},
                three_tags_ids[1]: {fake_processor_id, another_fake_processor_id},
                three_tags_ids[2]: {another_fake_processor_id},
            },
        )

    @pytest.mark.asyncio
    async def test_some_relations_exist(
        self,
        cataloged_entry: Entry,
        another_cataloged_entry: Entry,
        fake_processor_id: int,
        another_fake_processor_id: int,
        three_tags_ids: tuple[TagId, TagId, TagId],
    ) -> None:
        async with transaction() as trx:
            await apply_tags(
                trx, entry_id=cataloged_entry.id, processor_id=fake_processor_id, tags_ids=three_tags_ids[:2]
            )

        async with transaction() as trx:
            await apply_tags(
                trx, entry_id=cataloged_entry.id, processor_id=another_fake_processor_id, tags_ids=three_tags_ids[1:]
            )

        async with transaction() as trx:
            await apply_tags(
                trx, entry_id=another_cataloged_entry.id, processor_id=fake_processor_id, tags_ids=[three_tags_ids[1]]
            )

        async with (
            TableSizeDelta("o_relations_processors", delta=3),
            TableSizeDelta("o_relations", delta=2),
            TableSizeNotChanged("o_tags"),
        ):
            async with transaction() as trx:
                await tech_copy_relations(trx, cataloged_entry.id, another_cataloged_entry.id)

        await assert_has_tags({another_cataloged_entry.id: set(three_tags_ids)})

        await assert_tags_processors(
            entry_id=another_cataloged_entry.id,
            tag_processors={
                three_tags_ids[0]: {fake_processor_id},
                three_tags_ids[1]: {fake_processor_id, another_fake_processor_id},
                three_tags_ids[2]: {another_fake_processor_id},
            },
        )


class TestRemoveRelationsForEntries:
    @pytest.mark.asyncio
    async def test_nothing_to_remove(self, cataloged_entry: Entry, another_cataloged_entry: Entry) -> None:
        async with TableSizeNotChanged("o_relations"):
            await remove_relations_for_entries(execute, [cataloged_entry.id, another_cataloged_entry.id])

    @pytest.mark.asyncio
    async def test_success(
        self,
        cataloged_entry: Entry,
        another_cataloged_entry: Entry,
        fake_processor_id: int,
        another_fake_processor_id: int,
        three_tags_ids: tuple[TagId, TagId, TagId],
    ) -> None:
        async with transaction() as trx:
            await apply_tags(
                trx, entry_id=cataloged_entry.id, processor_id=fake_processor_id, tags_ids=three_tags_ids[:2]
            )

        async with transaction() as trx:
            await apply_tags(
                trx, entry_id=cataloged_entry.id, processor_id=another_fake_processor_id, tags_ids=three_tags_ids[1:]
            )

        async with transaction() as trx:
            await apply_tags(
                trx, entry_id=another_cataloged_entry.id, processor_id=fake_processor_id, tags_ids=three_tags_ids
            )

        async with transaction() as trx:
            await apply_tags(
                trx,
                entry_id=another_cataloged_entry.id,
                processor_id=another_fake_processor_id,
                tags_ids=three_tags_ids,
            )

        async with (
            TableSizeDelta("o_relations_processors", delta=-4),
            TableSizeDelta("o_relations", delta=-3),
            TableSizeNotChanged("o_tags"),
        ):
            async with transaction() as trx:
                await remove_relations_for_entries(trx, [cataloged_entry.id])

        await assert_has_tags({cataloged_entry.id: set(), another_cataloged_entry.id: set(three_tags_ids)})

        await assert_tags_processors(entry_id=cataloged_entry.id, tag_processors={})

        await assert_tags_processors(
            entry_id=another_cataloged_entry.id,
            tag_processors={
                three_tags_ids[0]: {fake_processor_id, another_fake_processor_id},
                three_tags_ids[1]: {fake_processor_id, another_fake_processor_id},
                three_tags_ids[2]: {fake_processor_id, another_fake_processor_id},
            },
        )


class TestApplyTagsProperties:

    @pytest.mark.asyncio
    async def test_no_properties(self) -> None:
        async with TableSizeNotChanged("o_tags_properties"):
            await apply_tags_properties(execute, [])

    @pytest.mark.asyncio
    async def test_first_time_save(
        self,
        three_tags_by_uids: dict[TagUid, TagId],
        three_processor_tags: tuple[ProcessorTag, ProcessorTag, ProcessorTag],
        fake_processor_id: int,
    ) -> None:

        properties = []

        for tag in three_processor_tags:
            tag.link = f"https://example.com?{tag.raw_uid}"
            properties.extend(
                tag.build_properties_for(
                    tag_id=three_tags_by_uids[tag.raw_uid], processor_id=fake_processor_id  # type: ignore
                )
            )

        async with TableSizeDelta("o_tags_properties", delta=3):
            await apply_tags_properties(execute, properties)

        loaded_tags_properties = await get_tags_properties(three_tags_by_uids.values())

        loaded_tags_properties.sort(key=lambda x: x.tag_id)
        properties.sort(key=lambda x: x.tag_id)

        assert loaded_tags_properties == properties

    @pytest.mark.asyncio
    async def test_save_duplicated(
        self,
        three_tags_by_uids: dict[TagUid, TagId],
        three_processor_tags: tuple[ProcessorTag, ProcessorTag, ProcessorTag],
        fake_processor_id: int,
    ) -> None:

        properties = []

        for tag in three_processor_tags:
            tag.link = f"https://example.com?{tag.raw_uid}"
            properties.extend(
                tag.build_properties_for(
                    tag_id=three_tags_by_uids[tag.raw_uid], processor_id=fake_processor_id  # type: ignore
                )
            )

        async with TableSizeDelta("o_tags_properties", delta=3):
            await apply_tags_properties(execute, properties)

        changed_properties = [
            properties[0].replace(),
            properties[1].replace(value="https://example.com?another-uid"),
            properties[2].replace(),
        ]

        async with TableSizeNotChanged("o_tags_properties"):
            await apply_tags_properties(execute, changed_properties)

        loaded_tags_properties = await get_tags_properties(three_tags_by_uids.values())

        loaded_tags_properties.sort(key=lambda x: x.tag_id)
        properties.sort(key=lambda x: x.tag_id)

        assert loaded_tags_properties != properties
        assert loaded_tags_properties == changed_properties

    @pytest.mark.asyncio
    async def test_error_on_duplicated_properties(
        self,
        three_tags_by_uids: dict[TagUid, TagId],
        three_processor_tags: tuple[ProcessorTag, ProcessorTag, ProcessorTag],
        fake_processor_id: int,
    ) -> None:

        properties = []

        for tag in three_processor_tags:
            tag.link = f"https://example.com?{tag.raw_uid}"
            properties.extend(
                tag.build_properties_for(
                    tag_id=three_tags_by_uids[tag.raw_uid], processor_id=fake_processor_id  # type: ignore
                )
            )

        properties.append(properties[0])

        async with TableSizeNotChanged("o_tags_properties"):
            with pytest.raises(errors.DuplicatedTagPropeties):
                await apply_tags_properties(execute, properties)


class TestCountTotalTags:
    @pytest.mark.asyncio
    async def test_no_tags(self) -> None:
        async with Delta(count_total_tags, delta=3):
            for _ in range(3):
                await get_or_create_id_by_tag(uuid.uuid4().hex)  # type: ignore


class TestCountTotalTagsPerCategory:

    @pytest.mark.asyncio
    async def test(self, cataloged_entry: Entry) -> None:
        tags = [uuid.uuid4().hex for _ in range(5)]

        processor_tags = [
            ProcessorTag(raw_uid=tags[0]),
            ProcessorTag(raw_uid=tags[1], categories={TagCategory.network_domain}),
            ProcessorTag(raw_uid=tags[2], categories={TagCategory.feed_tag}),
            ProcessorTag(raw_uid=tags[3], categories={TagCategory.network_domain, TagCategory.feed_tag}),
            ProcessorTag(raw_uid=tags[4], categories={TagCategory.feed_tag}),
        ]

        numbers_before = await count_total_tags_per_category()

        await apply_tags_to_entry(cataloged_entry.id, 1, processor_tags)

        numbers_after = await count_total_tags_per_category()

        assert numbers_after[TagCategory.network_domain] == numbers_before[TagCategory.network_domain] + 2
        assert numbers_after[TagCategory.feed_tag] == numbers_before[TagCategory.feed_tag] + 3


class TestCountTotalTagsPerType:

    @pytest.mark.asyncio
    async def test(self, cataloged_entry: Entry) -> None:
        tags = [uuid.uuid4().hex for _ in range(5)]

        processor_tags = [
            ProcessorTag(raw_uid=tags[0]),
            ProcessorTag(raw_uid=tags[1], link="https://example.com"),
            ProcessorTag(raw_uid=tags[2], link="https://example.com", categories={TagCategory.network_domain}),
            ProcessorTag(raw_uid=tags[3], link="https://example.com", categories={TagCategory.feed_tag}),
            ProcessorTag(raw_uid=tags[4], link="https://example.com"),
        ]

        numbers_before = await count_total_tags_per_type()

        await apply_tags_to_entry(cataloged_entry.id, 1, processor_tags)

        numbers_after = await count_total_tags_per_type()

        assert numbers_after[TagPropertyType.link] == numbers_before.get(TagPropertyType.link, 0) + 4
        assert numbers_after[TagPropertyType.categories] == numbers_before.get(TagPropertyType.categories, 0) + 2


class TestGetTagsForEntries:
    """Tested in other tests & code."""



import pydantic_settings

from ffun.core.settings import BaseSettings


class Settings(BaseSettings):
    max_returned_entries: int = 10000
    max_feeds_suggestions_for_site: int = 100
    max_entries_suggestions_for_site: int = 3
    max_entries_details_requests: int = 100

    news_outside_period: int = 100

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="FFUN_API_")


settings = Settings()
